#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/uaccess.h>
#include <linux/sysctl.h>

static int log_enabled = 1;
static int log_count = 0;
static int disabled_count = 0;
static int min = 0;
static int max = 1;
static struct ctl_table_header* header;

static int disabled_log = 0;
module_param(disabled_log, int, S_IRUSR | S_IWUSR);
MODULE_PARM_DESC(disabled_log, "Report when trying to log when disabled");

static int debug_log = 0;
module_param(debug_log, int, S_IRUSR | S_IWUSR);
MODULE_PARM_DESC(debug_log, "Output debug messages and message numbers");

// /proc/sys/kernel/kernlog_enabled: 1 to enable, any other value to disable
// /proc/sys/kernel/kernlog_count: total count of messages logged (read-only)
// /proc/sys/kernel/kernlog_disabled_count: total count of attempted logs while disabled

static struct ctl_table child_ctl_table[] = {
  {
    .procname = "kernlog_enabled",
    .data = &log_enabled,
    .maxlen = sizeof(int),
    .mode = 0644,
    .proc_handler = &proc_dointvec_minmax,
    .extra1 = &min,
    .extra2 = &max
  },
  {
    .procname = "kernlog_count",
    .data = &log_count,
    .maxlen = sizeof(int),
    .mode = 0444,
    .proc_handler = &proc_dointvec_minmax
  },
  {
    .procname = "kernlog_disabled_count",
    .data = &disabled_count,
    .maxlen = sizeof(int),
    .mode = 0444,
    .proc_handler = &proc_dointvec_minmax
  },
  {}
};

static struct ctl_table parent_ctl_table[] = {
  {
    .procname = "kernel",
    .mode = 0555,
    .child = child_ctl_table
  },
  {}
};

static ssize_t write_proc(struct file *filep, const char *buffer, size_t count, loff_t *offsetp)
{
    char string[256];
    count = count < 255 ? count : 255;

    if(copy_from_user(string, buffer, count) != 0) {
        pr_err("kernlog: Error occurred while getting message to log");
        return -EFAULT;
    }

    if (log_enabled == 1) {
        string[count] = '\0';
	log_count++;
        if (debug_log == 1) {
            printk("kernlog[%i]: %s", log_count, string);
        } else {
            printk(string);
        }
    } else {
        disabled_count++;
        if (disabled_log == 1) {
            pr_info("kernlog: Attempted to log message when disabled, disabled log count: %i", disabled_count);
        }
        return -EINVAL;
    }
    return count;
}

static ssize_t read_proc(struct file *file, char __user *ubuf,size_t count, loff_t *ppos)
{
    char buf[128];
    int len = 0;
    if (*ppos > 0 || count < 128) {
        return 0;
    }
    if (log_enabled == 1) {
        len += sprintf(buf, "Reading this file does not do anything. To log a message, write to this file.\n");
    } else {
        len += sprintf(buf, "Kernel logging is currently disabled. To enable, echo 1 > /proc/sys/kernel/kernlog_enabled.\n");
    }
    if (copy_to_user(ubuf, buf, len)) {
        return -EFAULT;
    }
    *ppos = len;
    return len;
}

static struct proc_ops proc_fops = {
    .proc_write = write_proc,
    .proc_read = read_proc,
};

static int proc_init(void) {
    struct proc_dir_entry *proc_file;
    proc_file = proc_create("kernlog", 0666, NULL, &proc_fops);

    if (proc_file == NULL) {
        pr_err("kernlog: Error while creating proc file");
        return -ENOMEM;
    }

    if (debug_log == 1) {
        pr_debug("kernlog: proc file created");
    }

    if (!(header = register_sysctl_table(parent_ctl_table))) {
        pr_err("kernlog: Error: Failed to register parent_ctl_table");
        return -EFAULT;
    }

    if (debug_log == 1) {
        pr_debug("kernlog: sysctl parameters created");
    }

    pr_info("kernlog: Kernel log module started");
    return 0;
}

static void proc_cleanup(void) {
    pr_info("kernlog: Kernel log module exiting");
    remove_proc_entry("kernlog", NULL);
    if (debug_log == 1) {
        pr_debug("kernlog: proc file removed");
    }

    unregister_sysctl_table(header);
    if (debug_log == 1) {
        pr_debug("kernlog: sysctl parameters removed");
    }
    pr_info("kernlog: Kernel log module exited");
}

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Log to the kernel log (dmesg) using /proc/kernlog");
module_init(proc_init);
module_exit(proc_cleanup);
